@@ -76,6 +76,7 @@ sub run_hook {
 	    $$textref .= "Bug-Ubuntu: https://bugs.launchpad.net/bugs/$bug\n";
 	}
     } elsif ($hook eq 'update-buildflags') {
+        $self->set_build_features(@params);
         $self->_add_build_flags(@params);
     } elsif ($hook eq 'builtin-system-build-paths') {
         return qw(/build/);
@@ -86,12 +87,14 @@ sub run_hook {
         umask 0022;
         # Reset locale to a sane default.
         $ENV{LC_COLLATE} = 'C.UTF-8';
+    } elsif ($hook eq 'backport-version-regex') {
+        return qr/~(bpo|deb)/;
     } else {
         return $self->SUPER::run_hook($hook, @params);
     }
 }
 
-sub _add_build_flags {
+sub set_build_features {
     my ($self, $flags) = @_;
 
     # Default feature states.
@@ -108,6 +111,9 @@ sub _add_build_flags {
             fixfilepath => 1,
             fixdebugpath => 1,
         },
+        optimize => {
+            lto => 0,
+        },
         sanitize => {
             address => 0,
             thread => 0,
@@ -156,83 +162,27 @@ sub _add_build_flags {
         ($abi, $os, $cpu) = ('', '', '');
     }
 
-    ## Global defaults
-
-    my $default_flags;
-    my $default_d_flags;
-    if ($opts_build->has('noopt')) {
-        $default_flags = '-g -O0';
-        $default_d_flags = '-fdebug';
-    } else {
-        $default_flags = '-g -O2';
-        $default_d_flags = '-frelease';
-    }
-    $flags->append('CFLAGS', $default_flags);
-    $flags->append('CXXFLAGS', $default_flags);
-    $flags->append('OBJCFLAGS', $default_flags);
-    $flags->append('OBJCXXFLAGS', $default_flags);
-    $flags->append('FFLAGS', $default_flags);
-    $flags->append('FCFLAGS', $default_flags);
-    $flags->append('GCJFLAGS', $default_flags);
-    $flags->append('DFLAGS', $default_d_flags);
-
     ## Area: future
 
     if ($use_feature{future}{lfs}) {
         my ($abi_bits, $abi_endian) = Dpkg::Arch::debarch_to_abiattrs($arch);
         my $cpu_bits = Dpkg::Arch::debarch_to_cpubits($arch);
 
-        if ($abi_bits == 32 and $cpu_bits == 32) {
-            $flags->append('CPPFLAGS',
-                           '-D_LARGEFILE_SOURCE -D_FILE_OFFSET_BITS=64');
+        if ($abi_bits != 32 or $cpu_bits != 32) {
+            $use_feature{future}{lfs} = 0;
         }
     }
 
-    ## Area: qa
-
-    # Warnings that detect actual bugs.
-    if ($use_feature{qa}{bug}) {
-        # C flags
-        my @cflags = qw(
-            implicit-function-declaration
-        );
-        foreach my $warnflag (@cflags) {
-            $flags->append('CFLAGS', "-Werror=$warnflag");
-        }
-
-        # C/C++ flags
-        my @cfamilyflags = qw(
-            array-bounds
-            clobbered
-            volatile-register-var
-        );
-        foreach my $warnflag (@cfamilyflags) {
-            $flags->append('CFLAGS', "-Werror=$warnflag");
-            $flags->append('CXXFLAGS', "-Werror=$warnflag");
-        }
-    }
-
-    # Inject dummy canary options to detect issues with build flag propagation.
-    if ($use_feature{qa}{canary}) {
-        require Digest::MD5;
-        my $id = Digest::MD5::md5_hex(int rand 4096);
-
-        foreach my $flag (qw(CPPFLAGS CFLAGS OBJCFLAGS CXXFLAGS OBJCXXFLAGS)) {
-            $flags->append($flag, "-D__DEB_CANARY_${flag}_${id}__");
-        }
-        $flags->append('LDFLAGS', "-Wl,-z,deb-canary-${id}");
-    }
-
     ## Area: reproducible
 
-    my $build_path;
-
     # Mask features that might have an unsafe usage.
     if ($use_feature{reproducible}{fixfilepath} or
         $use_feature{reproducible}{fixdebugpath}) {
         require Cwd;
 
-        $build_path = $ENV{DEB_BUILD_PATH} || Cwd::getcwd();
+        my $build_path =$ENV{DEB_BUILD_PATH} || Cwd::getcwd();
+
+        $flags->set_option_value('build-path', $build_path);
 
         # If we have any unsafe character in the path, disable the flag,
         # so that we do not need to worry about escaping the characters
@@ -243,31 +193,12 @@ sub _add_build_flags {
         }
     }
 
-    # Warn when the __TIME__, __DATE__ and __TIMESTAMP__ macros are used.
-    if ($use_feature{reproducible}{timeless}) {
-       $flags->append('CPPFLAGS', '-Wdate-time');
-    }
+    ## Area: optimize
 
-    # Avoid storing the build path in the binaries.
-    if ($use_feature{reproducible}{fixfilepath} or
-        $use_feature{reproducible}{fixdebugpath}) {
-        my $map;
-
-        # -ffile-prefix-map is a superset of -fdebug-prefix-map, prefer it
-        # if both are set.
-        if ($use_feature{reproducible}{fixfilepath}) {
-            $map = '-ffile-prefix-map=' . $build_path . '=.';
-        } else {
-            $map = '-fdebug-prefix-map=' . $build_path . '=.';
-        }
-
-        $flags->append('CFLAGS', $map);
-        $flags->append('CXXFLAGS', $map);
-        $flags->append('OBJCFLAGS', $map);
-        $flags->append('OBJCXXFLAGS', $map);
-        $flags->append('FFLAGS', $map);
-        $flags->append('FCFLAGS', $map);
-        $flags->append('GCJFLAGS', $map);
+    if ($opts_build->has('noopt')) {
+        $flags->set_option_value('optimize-level', 0);
+    } else {
+        $flags->set_option_value('optimize-level', 2);
     }
 
     ## Area: sanitize
@@ -283,31 +214,6 @@ sub _add_build_flags {
         $use_feature{sanitize}{leak} = 0;
     }
 
-    if ($use_feature{sanitize}{address}) {
-        my $flag = '-fsanitize=address -fno-omit-frame-pointer';
-        $flags->append('CFLAGS', $flag);
-        $flags->append('CXXFLAGS', $flag);
-        $flags->append('LDFLAGS', '-fsanitize=address');
-    }
-
-    if ($use_feature{sanitize}{thread}) {
-        my $flag = '-fsanitize=thread';
-        $flags->append('CFLAGS', $flag);
-        $flags->append('CXXFLAGS', $flag);
-        $flags->append('LDFLAGS', $flag);
-    }
-
-    if ($use_feature{sanitize}{leak}) {
-        $flags->append('LDFLAGS', '-fsanitize=leak');
-    }
-
-    if ($use_feature{sanitize}{undefined}) {
-        my $flag = '-fsanitize=undefined';
-        $flags->append('CFLAGS', $flag);
-        $flags->append('CXXFLAGS', $flag);
-        $flags->append('LDFLAGS', $flag);
-    }
-
     ## Area: hardening
 
     # Mask builtin features that are not enabled by default in the compiler.
@@ -356,7 +262,7 @@ sub _add_build_flags {
     }
 
     # Mask features that might be influenced by other flags.
-    if ($opts_build->has('noopt')) {
+    if ($flags->get_option_value('optimize-level') == 0) {
       # glibc 2.16 and later warn when using -O0 and _FORTIFY_SOURCE.
       $use_feature{hardening}{fortify} = 0;
     }
@@ -372,84 +278,12 @@ sub _add_build_flags {
 	$use_feature{hardening}{stackprotectorstrong} = 0;
     }
 
-    # PIE
-    if (defined $use_feature{hardening}{pie} and
-        $use_feature{hardening}{pie} and
-        not $builtin_feature{hardening}{pie}) {
-	my $flag = "-specs=$Dpkg::DATADIR/pie-compile.specs";
-	$flags->append('CFLAGS', $flag);
-	$flags->append('OBJCFLAGS',  $flag);
-	$flags->append('OBJCXXFLAGS', $flag);
-	$flags->append('FFLAGS', $flag);
-	$flags->append('FCFLAGS', $flag);
-	$flags->append('CXXFLAGS', $flag);
-	$flags->append('GCJFLAGS', $flag);
-	$flags->append('LDFLAGS', "-specs=$Dpkg::DATADIR/pie-link.specs");
-    } elsif (defined $use_feature{hardening}{pie} and
-             not $use_feature{hardening}{pie} and
-             $builtin_feature{hardening}{pie}) {
-	my $flag = "-specs=$Dpkg::DATADIR/no-pie-compile.specs";
-	$flags->append('CFLAGS', $flag);
-	$flags->append('OBJCFLAGS',  $flag);
-	$flags->append('OBJCXXFLAGS', $flag);
-	$flags->append('FFLAGS', $flag);
-	$flags->append('FCFLAGS', $flag);
-	$flags->append('CXXFLAGS', $flag);
-	$flags->append('GCJFLAGS', $flag);
-	$flags->append('LDFLAGS', "-specs=$Dpkg::DATADIR/no-pie-link.specs");
-    }
-
-    # Stack protector
-    if ($use_feature{hardening}{stackprotectorstrong}) {
-	my $flag = '-fstack-protector-strong';
-	$flags->append('CFLAGS', $flag);
-	$flags->append('OBJCFLAGS', $flag);
-	$flags->append('OBJCXXFLAGS', $flag);
-	$flags->append('FFLAGS', $flag);
-	$flags->append('FCFLAGS', $flag);
-	$flags->append('CXXFLAGS', $flag);
-	$flags->append('GCJFLAGS', $flag);
-    } elsif ($use_feature{hardening}{stackprotector}) {
-	my $flag = '-fstack-protector --param=ssp-buffer-size=4';
-	$flags->append('CFLAGS', $flag);
-	$flags->append('OBJCFLAGS', $flag);
-	$flags->append('OBJCXXFLAGS', $flag);
-	$flags->append('FFLAGS', $flag);
-	$flags->append('FCFLAGS', $flag);
-	$flags->append('CXXFLAGS', $flag);
-	$flags->append('GCJFLAGS', $flag);
-    }
-
-    # Fortify Source
-    if ($use_feature{hardening}{fortify}) {
-	$flags->append('CPPFLAGS', '-D_FORTIFY_SOURCE=2');
-    }
-
-    # Format Security
-    if ($use_feature{hardening}{format}) {
-	my $flag = '-Wformat -Werror=format-security';
-	$flags->append('CFLAGS', $flag);
-	$flags->append('CXXFLAGS', $flag);
-	$flags->append('OBJCFLAGS', $flag);
-	$flags->append('OBJCXXFLAGS', $flag);
-    }
-
-    # Read-only Relocations
-    if ($use_feature{hardening}{relro}) {
-	$flags->append('LDFLAGS', '-Wl,-z,relro');
-    }
-
-    # Bindnow
-    if ($use_feature{hardening}{bindnow}) {
-	$flags->append('LDFLAGS', '-Wl,-z,now');
-    }
-
     ## Commit
 
     # Set used features to their builtin setting if unset.
     foreach my $area (sort keys %builtin_feature) {
-        foreach my $feature (keys %{$builtin_feature{$area}}) {
-            $use_feature{$area}{$feature} //= $builtin_feature{$area}{$feature};
+        while (my ($feature, $enabled) = each %{$builtin_feature{$area}}) {
+            $flags->set_builtin($area, $feature, $enabled);
         }
     }
 
@@ -461,6 +295,184 @@ sub _add_build_flags {
     }
 }
 
+sub _add_build_flags {
+    my ($self, $flags) = @_;
+
+    ## Global default flags
+
+    my @compile_flags = qw(
+        CFLAGS
+        CXXFLAGS
+        OBJCFLAGS
+        OBJCXXFLAGS
+        FFLAGS
+        FCFLAGS
+        GCJFLAGS
+    );
+
+    my $default_flags;
+    my $default_d_flags;
+
+    my $optimize_level = $flags->get_option_value('optimize-level');
+    $default_flags = "-g -O$optimize_level";
+    if ($optimize_level == 0) {
+        $default_d_flags = '-fdebug';
+    } else {
+        $default_d_flags = '-frelease';
+    }
+
+    $flags->append($_, $default_flags) foreach @compile_flags;
+    $flags->append('DFLAGS', $default_d_flags);
+
+    ## Area: future
+
+    if ($flags->use_feature('future', 'lfs')) {
+        $flags->append('CPPFLAGS',
+                       '-D_LARGEFILE_SOURCE -D_FILE_OFFSET_BITS=64');
+    }
+
+    ## Area: qa
+
+    # Warnings that detect actual bugs.
+    if ($flags->use_feature('qa', 'bug')) {
+        # C flags
+        my @cflags = qw(
+            implicit-function-declaration
+        );
+        foreach my $warnflag (@cflags) {
+            $flags->append('CFLAGS', "-Werror=$warnflag");
+        }
+
+        # C/C++ flags
+        my @cfamilyflags = qw(
+            array-bounds
+            clobbered
+            volatile-register-var
+        );
+        foreach my $warnflag (@cfamilyflags) {
+            $flags->append('CFLAGS', "-Werror=$warnflag");
+            $flags->append('CXXFLAGS', "-Werror=$warnflag");
+        }
+    }
+
+    # Inject dummy canary options to detect issues with build flag propagation.
+    if ($flags->use_feature('qa', 'canary')) {
+        require Digest::MD5;
+        my $id = Digest::MD5::md5_hex(int rand 4096);
+
+        foreach my $flag (qw(CPPFLAGS CFLAGS OBJCFLAGS CXXFLAGS OBJCXXFLAGS)) {
+            $flags->append($flag, "-D__DEB_CANARY_${flag}_${id}__");
+        }
+        $flags->append('LDFLAGS', "-Wl,-z,deb-canary-${id}");
+    }
+
+    ## Area: reproducible
+
+    # Warn when the __TIME__, __DATE__ and __TIMESTAMP__ macros are used.
+    if ($flags->use_feature('reproducible', 'timeless')) {
+       $flags->append('CPPFLAGS', '-Wdate-time');
+    }
+
+    # Avoid storing the build path in the binaries.
+    if ($flags->use_feature('reproducible', 'fixfilepath') or
+        $flags->use_feature('reproducible', 'fixdebugpath')) {
+        my $build_path = $flags->get_option_value('build-path');
+        my $map;
+
+        # -ffile-prefix-map is a superset of -fdebug-prefix-map, prefer it
+        # if both are set.
+        if ($flags->use_feature('reproducible', 'fixfilepath')) {
+            $map = '-ffile-prefix-map=' . $build_path . '=.';
+        } else {
+            $map = '-fdebug-prefix-map=' . $build_path . '=.';
+        }
+
+        $flags->append($_, $map) foreach @compile_flags;
+    }
+
+    ## Area: optimize
+
+    if ($flags->use_feature('optimize', 'lto')) {
+        my $flag = '-flto=auto -ffat-lto-objects';
+        $flags->append($_, $flag) foreach (@compile_flags, 'LDFLAGS');
+    }
+
+    ## Area: sanitize
+
+    if ($flags->use_feature('sanitize', 'address')) {
+        my $flag = '-fsanitize=address -fno-omit-frame-pointer';
+        $flags->append('CFLAGS', $flag);
+        $flags->append('CXXFLAGS', $flag);
+        $flags->append('LDFLAGS', '-fsanitize=address');
+    }
+
+    if ($flags->use_feature('sanitize', 'thread')) {
+        my $flag = '-fsanitize=thread';
+        $flags->append('CFLAGS', $flag);
+        $flags->append('CXXFLAGS', $flag);
+        $flags->append('LDFLAGS', $flag);
+    }
+
+    if ($flags->use_feature('sanitize', 'leak')) {
+        $flags->append('LDFLAGS', '-fsanitize=leak');
+    }
+
+    if ($flags->use_feature('sanitize', 'undefined')) {
+        my $flag = '-fsanitize=undefined';
+        $flags->append('CFLAGS', $flag);
+        $flags->append('CXXFLAGS', $flag);
+        $flags->append('LDFLAGS', $flag);
+    }
+
+    ## Area: hardening
+
+    # PIE
+    my $use_pie = $flags->use_feature('hardening', 'pie');
+    my %hardening_builtins = $flags->get_builtins('hardening');
+    if (defined $use_pie && $use_pie && ! $hardening_builtins{pie}) {
+	my $flag = "-specs=$Dpkg::DATADIR/pie-compile.specs";
+        $flags->append($_, $flag) foreach @compile_flags;
+	$flags->append('LDFLAGS', "-specs=$Dpkg::DATADIR/pie-link.specs");
+    } elsif (defined $use_pie && ! $use_pie && $hardening_builtins{pie}) {
+	my $flag = "-specs=$Dpkg::DATADIR/no-pie-compile.specs";
+        $flags->append($_, $flag) foreach @compile_flags;
+	$flags->append('LDFLAGS', "-specs=$Dpkg::DATADIR/no-pie-link.specs");
+    }
+
+    # Stack protector
+    if ($flags->use_feature('hardening', 'stackprotectorstrong')) {
+	my $flag = '-fstack-protector-strong';
+        $flags->append($_, $flag) foreach @compile_flags;
+    } elsif ($flags->use_feature('hardening', 'stackprotector')) {
+	my $flag = '-fstack-protector --param=ssp-buffer-size=4';
+        $flags->append($_, $flag) foreach @compile_flags;
+    }
+
+    # Fortify Source
+    if ($flags->use_feature('hardening', 'fortify')) {
+	$flags->append('CPPFLAGS', '-D_FORTIFY_SOURCE=2');
+    }
+
+    # Format Security
+    if ($flags->use_feature('hardening', 'format')) {
+	my $flag = '-Wformat -Werror=format-security';
+	$flags->append('CFLAGS', $flag);
+	$flags->append('CXXFLAGS', $flag);
+	$flags->append('OBJCFLAGS', $flag);
+	$flags->append('OBJCXXFLAGS', $flag);
+    }
+
+    # Read-only Relocations
+    if ($flags->use_feature('hardening', 'relro')) {
+	$flags->append('LDFLAGS', '-Wl,-z,relro');
+    }
+
+    # Bindnow
+    if ($flags->use_feature('hardening', 'bindnow')) {
+	$flags->append('LDFLAGS', '-Wl,-z,now');
+    }
+}
+
 sub _build_tainted_by {
     my $self = shift;
     my %tainted;
